package week09

import scalafx.scene.paint.Color

import scala.util.Random

/**
 * A Shape trait alakzatot tárol a 2D síkon, öt metódusa van, alapértelmezett ??? implementációval, melyet a
 * két implementáló osztály, a Kor és a Teglalap felül kell írjanak.
 */
trait Shape {
  /** contains: igazat kell adjon, ha az (x,y) koordinátájú pont ill. a p Point
   * benne van az alakzatban vagy annak határán.
   */
  def contains(x: Double, y: Double): Boolean = contains(Point(x, y))

  def contains(p: Point): Boolean

  /** zoom: az origóból nézve skálázza az alakzatot ratio-szorosára.
   */
  def zoom(ratio: Double): Shape

  /**
   * összeadás: eltolja az alakzatot az (x,y) vektorral ill. a v Vektorral.
   */
  def +(x: Double, y: Double): Shape

  def +(v: Vektor): Shape = this.+(v.dx, v.dy)
}

/**
 * A Vektor osztály egy 2D vektort (dx, dy koordináta-különbségeket) tárol, ehhez nem kell nyúljunk.
 */
case class Vektor(dx: Double, dy: Double)

/**
 * A Point osztály egy 2D pontot tárol a síkon, két implementálandó metódussal:
 */
case class Point(x: Double, y: Double) {
  /**
   * Vektor és Point összeadása: toljuk el a pontot a vektorral
   */
  def +(diff: Vektor): Point = Point(x + diff.dx, y + diff.dy)

  /**
   * a to metódus adja vissza a két Point távolságát
   */
  def to(that: Point): Double =
    Math.sqrt(Math.pow(this.x - that.x, 2.0) + Math.pow(this.y - that.y, 2.0))
}

/**
 * Generally: Error -- very abnormal error, hard to recover
 * Exception -- might happen under "normal" circumstances
 * Assertion -- removed at runtime
 */

/**
 * Implementáljuk a Shape trait metódusait a Kor ill. Teglalap osztályokban!
 * Oldjuk meg azt is, hogy ha invalid paraméterekkel próbáljuk inicialiálni az alakzatokat
 * (a kört negatív sugárral, a téglalapot negatív szélességgel vagy magassággal),
 * akkor a vonatkozó paramétereket 50-re állítsa be inkább a program.
 *
 * Írjunk továbbá a körnek is és a téglalapnak is kétparaméteres konstruktort, mely r-t ill. width-et és height-et
 * szintén 50-re állítja.
 *
 * (Ilyesmihez a companion object apply metódusait kell felülírjuk.)
 */
case class Kor private(x: Double, y: Double, r: Double) extends Shape {
  /** contains: igazat kell adjon, ha az (x,y) koordinátájú pont ill. a p Point
   * benne van az alakzatban vagy annak határán.
   */
  //require( r >= 0.0 , "A kör sugara negatív")

  override def contains(p: Point): Boolean = (p to Point(x, y)) <= r

  /** zoom: az origóból nézve skálázza az alakzatot ratio-szorosára.
   */
  override def zoom(ratio: Double) = Kor(x * ratio, y * ratio, r * ratio)

  /**
   * összeadás: eltolja az alakzatot az (x,y) vektorral ill. a v Vektorral.
   */
  override def +(x: Double, y: Double) = Kor(this.x + x, this.y + y, r)
}

object Kor {
  // - a véletlen kör középpontjának koordinátái a -50 .. 50 intervallumba essenek,
  // sugara pedig a 25..50 intervallumba,
  def generateRandom = {
    val x = Random.nextInt(101) - 50
    val y = Random.nextInt(101) - 50
    val r = Random.nextInt(26) + 25
    Kor(x, y, r)
  }

  def apply(x: Double, y: Double, r: Double): Kor = {
    if (r >= 0) new Kor(x, y, r) else new Kor(x, y, 50)
  }

  def apply(x: Double, y: Double): Kor = Kor(x, y, 50)
}

case class Teglalap(x: Double, y: Double, width: Double, height: Double) extends Shape {
  override def contains(p: Point): Boolean =
    (this.x <= p.x) && (p.x <= this.x + this.width) &&
      (this.y <= p.y) && (p.y <= this.y + this.height)

  /** zoom: az origóból nézve skálázza az alakzatot ratio-szorosára.
   */
  override def zoom(ratio: Double) = Teglalap(x * ratio, y * ratio, width * ratio, height * ratio)

  /**
   * összeadás: eltolja az alakzatot az (x,y) vektorral ill. a v Vektorral.
   */
  override def +(x: Double, y: Double) = Teglalap(this.x + x, this.y + y, width, height)
}

object Teglalap {
  // - a véletlen téglalap bal felső sarkának koordinátái a -50..-10,
  // jobb alsó sarkának koordinátái a 10..50 intervallumba

  def generateRandom = {
    val x0 = Random.nextInt(41) - 50
    val y0 = Random.nextInt(41) - 50
    val x1 = Random.nextInt(41) + 10
    val y1 = Random.nextInt(41) + 10
    Teglalap(x0, y0, x1 - x0, y1 - y0)
  }

}

/**
 * Egy ColoredShape egy Shape alakzatot és két színt: a körvonal színét és a kitöltés színét tárolja.
 * Implementáljuk a metódusait!
 */
case class ColoredShape(shape: Shape, draw: Color, fill: Color) {
  /**
   * A contains metódus adja vissza, hogy a shape tartalmazza-e a megadott pontot;
   * a zoom ill. translate a shape-et transzformálja el,
   * a flip metódus pedig cserélje fel a körberajzoló és a kitöltő színeket!
   */
  def contains(x: Double, y: Double): Boolean = shape.contains(x, y)

  def contains(p: Point): Boolean = shape.contains(p)

  def zoom(ratio: Double): ColoredShape = ColoredShape(shape.zoom(ratio), draw, fill)

  def translate(x: Double, y: Double): ColoredShape = ColoredShape(shape + (x, y), draw, fill)

  def translate(v: Vektor): ColoredShape = ColoredShape(shape + v, draw, fill)

  def flip: ColoredShape = ColoredShape(shape, fill, draw)
}

/**
 * A ColoredShape companion objectben két metódust kell implementáljunk.
 */
object ColoredShape {
  private val random = Random
  private val colors = Vector(Color.Yellow, Color.Green, Color.Blue, Color.Purple, Color.Orange)

  /**
   * A generateColors metódus generáljon két KÜLÖNBÖZŐ színt a colors vektorból és adja vissza őket párként!
   */
  def generateColors: (Color, Color) = {
    val i = random.nextInt(colors.length)
    val j = random.nextInt(colors.length)
    if (i != j) (colors(i), colors(j)) else generateColors
  }

  /**
   * A generateRandom metódus 50-50% eséllyel egy véletlen kört vagy egy véletlen téglalapot adjon vissza,
   * két véletlen, eltérő színnel inicializálva.
   * - a véletlen kör középpontjának koordinátái a -50 .. 50 intervallumba essenek, sugara pedig a 25..50 intervallumba,
   * - a véletlen téglalap bal felső sarkának koordinátái a -50..-10, jobb alsó sarkának koordinátái a 10..50 intervallumba
   * essenek! A két alakzat generátor algoritmusait a Kor ill. Teglalap companion objectjébe írjuk,
   * generateRandom címmel.
   */
  def generateRandom: ColoredShape = {
    val shape = random.nextBoolean() match {
      case false => Kor.generateRandom
      case true => Teglalap.generateRandom
    }
    val (draw, fill) = generateColors
    ColoredShape(shape, draw, fill)
  }
}

/**
 * A Port osztály egy viewportot tárol, adott a középpontja és szélessége-magasságának a FELE
 * (tehát a center.x +- width és center.y +- height koordináták adják a sarkait).
 * Implementáljuk ennek is a metódusait!
 */
case class Port(center: Point, width: Double, height: Double) {
  def zoom(factor: Double): Port = Port(
    Point(center.x * factor, center.y * factor),
    width * factor, height * factor
  )

  def translate(diff: Vektor): Port = Port(center + diff, width, height)

  def contains(kor: Kor): Boolean = {
    if ((width >= kor.r) && (height >= kor.r))
      Teglalap(center.x - width + kor.r, center.y - height + kor.r, 2 * width - 2 * kor.r, 2 * height - 2 * kor.r)
        .contains(kor.x, kor.y)
    else false
  }

  def contains(teglalap: Teglalap): Boolean = {
    (center.x - width <= teglalap.x) &&
      (teglalap.x + teglalap.width <= center.x + width) &&
      (center.y - height <= teglalap.y) &&
      (teglalap.y + teglalap.height <= center.y + height)
  }

  def contains(shape: Shape): Boolean = {
    shape match {
      case kor: Kor => contains(kor)
      case teglalap: Teglalap => contains(teglalap)
      case _ => {
        System.err.println("Ismeretlen Shape (Port.contains): " + shape); false
      }
    }
  }

  def corners: List[Point] =
    for (i <- List(-1, 1); j <- List(-1, 1)) yield Point(center.x + i * width, center.y + j * height)
}

/**
 * az ActorVector egy viewportot és ColoredShape-ek egy vektorát tárolja.
 * Implementáljuk a metódusait!
 */
case class ActorVector(canvas: Port, shapes: Vector[ColoredShape]) {
  //a translate függvények tolják el a canvast is és minden shapet a vektorban
  def translate(x: Double, y: Double): ActorVector = translate(Vektor(x, y))

  def translate(v: Vektor): ActorVector =
    ActorVector(canvas.translate(v), shapes map {
      _.translate(v)
    })

  //a zoom függvény zoomolja a canvast is és minden shapet a vektorban
  def zoom(ratio: Double): ActorVector =
    ActorVector(canvas.zoom(ratio), shapes map {
      _.zoom(ratio)
    })

  //az addRandomShape adjon hozzá egy random színes alakzatot az ott implementált módszerrel
  def addRandomShape: ActorVector =
    ActorVector(canvas, shapes :+ ColoredShape.generateRandom)

  //a flipShapesAt metódus flippelje azoknak az alakzatoknak a színeit, melyek tartalmazzák az (x,y) pontot
  def flipShapesAt(x: Double, y: Double): ActorVector =
    ActorVector(canvas, shapes map { shape => if (shape.contains(x, y)) shape.flip else shape })

  //a removeShapesAt metódus törölje azoknak az alakzatokat a vektorból, melyek tartalmazzák az (x,y) pontot!
  def removeShapesAt(x: Double, y: Double): ActorVector =
    ActorVector(canvas, shapes filterNot {
      _.contains(x, y)
    })
}

object Gyakorlat extends App {
  val kor = Kor(1, 2, 10)
  val kor2 = Kor(1, 3)
  println(kor)
  println(kor2)
  println(System.getProperty("java.version"))
}
