package week10

import scala.collection.immutable.HashMap

/**
 * A mai ZH feladatban egy egyszerű számológépet fogunk implementálni, ami támogat
 * - összeadást, szorzást, kivonást, osztást
 * - mínusz egyszerest
 * - zárójelezést
 * - változókat, amikhez racionális értékeket rendelhetünk
 * - a cél: egy string-listában megadott utasítás-sorozatot felparsolni és kiértékelni, pl:
 * List( "x := 3", "y := 1/2", "x + y", "x := 1", "x + y" ) outputja List(Rational(7/2),Rational(3/2)) lenne.
 */

/**
 * Készítsünk egy racionális számokat tároló osztályt!
 * - adattagok: szamlalo, bármilyen Int lehet; nevezo: csak pozitív Int lehet.
 * Ha nem ilyen az input, dobjunk a konstruktorból egy IllegalArgumentException("Rational constuctor") -t.
 *
 * note: az osztály kiterjeszti a kifejezés traitet is és az Ordered[Rational]-t is, ezt így tehetjük meg.
 */
case class Rational(szamlalo: Int, nevezo: Int) extends Ordered[Rational] with Kifejezes{
  require(nevezo > 0 , "Rational constuctor") //ENH: ha nevezo < 0 => (-szamlalo, -nevezo) -- a Rational object-be, itt meg konstruktort priváttá
  /**
   * A négy bináris alapművelet értelemszerűen adja vissza a this + that, this - that, stb racionális számokat!
   * Ha a művelet nem értelmezett, akkor dobjon IllegalArgumentException("Rational: Invalid operation")-t.
   */
  def +(that: Rational): Rational = Rational(this.szamlalo * that.nevezo + this.nevezo * that.szamlalo, this.nevezo * that.nevezo)
  def -(that: Rational): Rational = Rational(this.szamlalo * that.nevezo - this.nevezo * that.szamlalo, this.nevezo * that.nevezo)
  def *(that: Rational): Rational = Rational(this.szamlalo * that.szamlalo, this.nevezo * that.nevezo)
  def /(that: Rational): Rational =
    if (that.szamlalo == 0 ) throw new IllegalArgumentException("Rational: Invalid operation")
    else if(that.szamlalo > 0 ) Rational(this.szamlalo * that.nevezo, this.nevezo * that.szamlalo)
    else Rational(-this.szamlalo * that.nevezo, -this.nevezo * that.szamlalo)
  /**
   * A unary_-() metódus adja vissza a szám ellentettjét! (pl Rational(-1,3).unary_-() = Rational(1,3). )
   *
   * A unary_-() metódus egyébként a "mínusz" operátort definiálja, innentől használhatjuk pl. a
   * val x = - Rational(1,3) -like értékadásokat.
   */
  def unary_-(): Rational = Rational(-szamlalo, nevezo)
  /** A compare metódus adjon
   * - negatív értéket, ha this < that,
   * - pozitívat, ha this > that,
   * - nullát, ha this == that!
   *
   * Mivel egyébként kiterjesztjük az Ordered[Rational] traitet, így ennek a metódusnak a definiálásával
   * kapunk egy <, >, <= és >= operátort is, melyek ezzel konzisztensek.
   */
  override def compare(that: Rational): Int = (this - that).szamlalo
  /**
   * Implementáljuk az örökölt equals metódust a következőképp:
   * - ha that egy Rational, akkor legyenek egyenlőek, ha ugyanaz az értékük, pl. Rational(1,2).equals(Rational(2,4)) == true
   * - ha that nem Rational, akkor legyen false.

   * (note: match kifejezés típusra az idiomatikus megoldás.)
   */
  override def equals(that: Any): Boolean = that match {
    case thatRat : Rational => (this compare thatRat)==0
    case _ => false
  }
  /**
   * Adja vissza szamlalo/nevezo alakban (szóközök nélkül). Nem baj, ha nincs egyszerűsítve!
   * Pl. Rational(3,6).toString == "3/6".
   */
  override def toString: String = szamlalo + "/" + nevezo

  def apply(v: ValtozoErtekadas): Option[Rational] = Some(this)
}

/**
 * Készítsünk egy változó-értékadásokat tartalmazó osztályt! A map adattagjában, melyet konstruktorban kap meg,
 * tárolja, hogy az adott nevű változó értéke éppen mennyi.
 */
case class ValtozoErtekadas(map: Map[String,Rational]) {
  /**
   * set: állítsuk be a varName nevű változó értékét value-ra!
   * Pl. ValtozoErtekadas().set("x",Rational(1,2)).set("y",Rational(1,3)).set("x",Rational(1,4)) után
   * az értékben tárolt értékadás mellett x értéke 1/4, y értéke 1/3, minden más változó értéke definiálatlan legyen.
   */
  def set(varName: String, value: Rational): ValtozoErtekadas =
    ValtozoErtekadas(map.updated(varName,value))

  /**
   * unset: felejtsük el a varName változó értékét.
   * Pl. Valtozoertekadas().set("x",Rational(1,2)).unset("y").set("y",Rational(1,3)).unset("x") után
   * az értékadás mellett y értéke 1/3, minden más definiálatlan.
   */
  def unset(varName: String): ValtozoErtekadas =
    ValtozoErtekadas(map.removed(varName))

  /**
   * Adjuk vissza Some(r)-t, ha varName-nek van beállított értéke és az r; None-t, ha definiálatlan az értéke!
   */
  def get(varName: String): Option[Rational] = map.get(varName)

  /**
   * Adjuk vissza, hogy be van-e állítva az értékadásban a varName változó értéke valamire (true)
   * vagy definiálatlan (false)!
   */
  def isSet(varName: String): Boolean = map.contains(varName)

  /**
   * A map függvény hajtsa végre a megadott f leképezést minden változó értékén!
   * Tehát pl. ha f az { x => x * x } függvény, és a változóértékadásunkban "x" értéke Rational(3,2),
   * "y" értéke pedig Rational(-1,2), akkor a map által visszaadott értékadásban "x" értéke Rational(9,4),
   * "y" értéke pedig Rational(1,4) lesz.
   */
  def map(f: Rational => Rational): ValtozoErtekadas =
    ValtozoErtekadas( map.map( entry => (entry._1, f(entry._2) ) ) )

  /**
   * A foreach függvény értékelje ki a megadott f(varName, value) kifejezést minden, az értékadásban szereplő
   * entryre! Írjuk meg úgy is, hogy párra működjön és kétváltozósra is.
   */
  def foreach[T](f: (String,Rational)=>T): Unit = map foreachEntry f
  def foreach[T](f: ((String,Rational)) => T): Unit = map foreach f

}
object ValtozoErtekadas {
  /**
   * A ValtozoErtekadas() hívás adjon vissza egy üres értékadást, amiben minden változó értéke definiálatlan.
   */
  def apply(): ValtozoErtekadas = ValtozoErtekadas(HashMap())
}

/**
 * A Kifejezes trait egy racionális számokból és változókból álló kifejezés-fát tárol,
 * pl. a (x+1/2)*(y-z) kifejezést egy
 * Mul(Add(Variable("x"),Rational(1,2)),Sub(Variable("y"),Variable("z"))) Kifejezes ábrázolja.
 *
 * Van egy apply(v: ValtozoErtekadas) metódusa, ami kiértékeli egy adott változó-értékadás mellett.
 */
trait Kifejezes {
  /**
   * Ha a v változó-értékadás beállít minden változót, ami a kifejezésben szerepel, akkor Some(r)-ként
   * jöjjön vissza a kifejezés értéke az értékadás mellett; ha van a kifejezésben olyan változó, mely az
   * értékadásban definiálatlan, vagy ha a kifejezés nem értékelhető ki (pl. mert van benne nullával
   * osztás), akkor None-ként jöjjön vissza!
   *
   * (Itt a traitben persze nem tudjuk implementálni ezt a függvényt, töröljük is ki a kérdőjeleket, aztán oldjuk
   * meg, hogy minden leforduljon!.)
   */
  def apply(v: ValtozoErtekadas): Option[Rational]
}

/**
 * A négy alapművelet osztályai. Implementáljuk őket! (az apply metódust muszáj lesz.)
 * (Gondolkodhatunk "szebb" megoldásban is, mint ami elsőre eszünkbe jut, de jó úgy is, ahogy elsőre eszünkbe jut)
 */
//összeadás
case class Add(left: Kifejezes, right: Kifejezes) extends Kifejezes {
  def apply(v: ValtozoErtekadas): Option[Rational] =
    for( r1 <- left(v); r2 <- right(v) ) yield r1+r2
}
//kivonás
case class Sub(left: Kifejezes, right: Kifejezes) extends Kifejezes{
  def apply(v: ValtozoErtekadas): Option[Rational] =
    for( r1 <- left(v); r2 <- right(v) ) yield r1-r2
}
//szorzás
case class Mul(left: Kifejezes, right: Kifejezes) extends Kifejezes{
  def apply(v: ValtozoErtekadas): Option[Rational] =
    for( r1 <- left(v); r2 <- right(v) ) yield r1*r2
}
//osztás
case class Div(left: Kifejezes, right: Kifejezes) extends Kifejezes{
  def apply(v: ValtozoErtekadas): Option[Rational] =
    for(
      r1 <- left(v);
      r2 <- right(v) if r2 != Rational(0,1))
      yield r1/r2
}
//ellentett, mínusz eggyel szorzás
case class Neg(kernel: Kifejezes) extends Kifejezes{
  def apply(v: ValtozoErtekadas): Option[Rational] = kernel(v).map{ - _ }
}
//változó, a nevével adott
case class Variable(varName: String) extends Kifejezes{
  def apply(v: ValtozoErtekadas): Option[Rational] = v.get(varName)
}
//és a Rational osztály is kiterjeszti a Kifejezes traitet, így annak is kell készítsünk egy megfelelő apply metódust.

/**
 * INNENTŐL PLUSZ PONTOS FELADATOK VANNAK
 * Plusz pontért ha ezzel kész vagyunk, az előzőeket használva próbáljunk meg befejezni egy parsert,
 * ami egy input stringből elkészít egy kifejezést, ha az egy valid expression és none-t, ha nem!
 */


/**
 * A KifejezesParser objektum célja: egy kifejezés-stringből (mint pl "(x  + 1/2) *y") egy Kifejezes objektumot
 * (mint Mul(Add(Variable("x"),Rational(1,2)),Variable("y")) ) készítsen.
 *
 * Ehhez tartalmaz néhány helper metódust is.
 */
object KifejezesParser {
  /**
   * A removeWhitespace metódus tartsa meg az input stringből csak
   * - az angol ábécé kisbetűit, a-z,
   * - a számjegyeket, 0-9,
   * - a műveleti jeleket, + - * /,
   * - és a zárójeleket: ( )
   * tartsa meg, a többit dobja el!
   * Pl. removeWhitespace("(x  + 1/2) *yá z") == "(x+1/2)*yz".
   */
  def removeWhitespace(input: String): String = ???

  /**
   * A computeNestingVector metódus számoljon ki egy Vector[Int]-et, amibe az input
   * (whitespace-mentes) string pozícióinak zárójelezési mélységét számolja ki!
   * (0-ról indul, '(' növeli, ')' csökkenti). A nyitójel a nagyobb, a csukójel már a kisebb mélységbe számítson.
   * Pl.
   * computeNestingVector("(x+(1/2))*yz") = Vector(1,1,1,2,2,2,2,1,0,0,0,0).
   */
  def computeNestingVector(input: String): Vector[Int] = ???

  /**
   * A canBeValid metódus mondja meg, hogy az input (whitespace-mentes) stringben a zárójelezés rendben van-e,
   * azaz
   * - minden pozíción nemnegatív,
   * - és az utolsó pozíció mélysége 0 -e!
   * Pl. canBeValid("(x+(1/2))*yz")==true, canBeValid("(+)(+)") == true, de canBeValid("())(") == false.
   */
  def canBeValid(input: String): Boolean = ???

  /**
   * A breakDeepestParen metódus az input stringet "vágja" három részre: a középsőben legyen a
   * legmélyebben lévő zárójelezett rész, de már a zárójelei nélkül, a két szélsőben meg az azt megelőző ill. követő
   * rész, változatlanul! Ha több, egyforma mélyen lévő zárójelezés-rész is van, akkor az első ilyet "vágjuk ki".
   * Ha nincs a stringben zárójel, dobjunk IllegalArgumentException -t.
   *
   * Pl. breakDeepestParen("(x+(1/2))*yz") = ("(x+", "1/2", ")*yz") , mert az "(1/2)" része a legmélyebben zárójelezett.
   */
  def breakDeepestParen(s: String): (String, String, String) = ???

  /**
   * A hasChar metódus adja vissza, hogy szerepel-e az input stringben a megadott karakter.
   */
  def hasChar(s: String, c: Char): Boolean = ???

  def parse(s: String): Option[Kifejezes] = ???
}

object ZH extends App{
  //írj magadnak ide teszteket próbálkozni if you wish
}
